@@ -1,3 +1,9 @@
+/* Copyright (c) 1998, 1999 John E. Davis
+ * This file is part of the S-Lang library.
+ *
+ * You may distribute under the terms of either the GNU General Public
+ * License or the Perl Artistic License.
+ */
 #include <stdio.h>
 
 #ifndef SLANG_VERSION
@@ -33,7 +39,7 @@ extern int SLcurses_wprintw (SLcurses_Window_Type *, char *, ...);
 extern SLcurses_Window_Type *SLcurses_newwin (unsigned int, unsigned int,
 					      unsigned int, unsigned int);
 
-extern SLcurses_Window_Type *SLcurses_subwin (SLcurses_Window_Type *, 
+extern SLcurses_Window_Type *SLcurses_subwin (SLcurses_Window_Type *,
 					      unsigned int, unsigned int,
 					      unsigned int, unsigned int);
 
@@ -61,7 +67,6 @@ extern int SLcurses_waddnstr (SLcurses_Window_Type *, char *, int);
   ((-1 == wmove((w),(y),(x))) ? -1 : waddch((w),(c)))
 #define mvaddch(y,x,c)         	mvwaddch(stdscr,(y),(x),(c))
 
-
 extern int SLcurses_wclear (SLcurses_Window_Type *w);
 extern int SLcurses_printw (char *, ...);
 
@@ -77,7 +82,7 @@ extern int SLcurses_scanw (char *, ...);
 #endif
 
 extern SLcurses_Window_Type *SLcurses_Stdscr;
-#define WINDOW SLcurses_Window_Type 
+#define WINDOW SLcurses_Window_Type
 #define stdscr SLcurses_Stdscr
 
 #define subwin		SLcurses_subwin
@@ -213,7 +218,6 @@ extern int SLcurses_Is_Endwin;
    (w)->scroll_max=(w)->nrows, \
    wscrl((w), -1))
 
-
 extern SLtt_Char_Type SLcurses_Acs_Map [128];
 #define acs_map SLcurses_Acs_Map
 
@@ -270,13 +274,13 @@ extern SLtt_Char_Type SLcurses_Acs_Map [128];
 #define COLOR_CYAN	SLSMG_COLOR_CYAN
 #define COLOR_WHITE	SLSMG_COLOR_LGRAY
 
-#define COLORS		8
-#define COLOR_PAIRS	(8 * 8)
+extern int SLcurses_Num_Colors;
+#define COLORS		SLcurses_Num_Colors
+#define COLOR_PAIRS	(SLcurses_Num_Colors*SLcurses_Num_Colors)
 
 #define init_pair(_x,_f,_b) \
  SLtt_set_color_object((_x), ((_f) == (_b) ? 0x0700 : ((_f) | ((_b) << 8)) << 8))
 
-
 #define scrollok(a,b) ((a)->scroll_ok = (b))
 #define getyx(a,y,x)  (y=(a)->_cury, x=(a)->_curx)
 #define getmaxyx(a,y,x)  (y=(a)->nrows, x=(a)->ncols)
@@ -323,7 +327,6 @@ extern int SLcurses_clearok (SLcurses_Window_Type *, int);
 #define trace(x) SLcurses_nil()
 #define tigetstr(x) NULL
 
-
 /* These have no place in C */
 #define TRUE 1
 #define FALSE 0
